////////////////////////////////////// c++ ////////////////////////////////////
//
//  Module           : HeapMin.hh
//  Description      :
//  Author           : Christoph Laux
//  Email            : claux@ix.urz.uni-heidelberg.de
//  Copyright        : University of Heidelberg
//  Created on       : Thu Febr 27 09:21:54 2003
//  Last modified by : laux
//  Last modified on : Thu Febr 27 09:21:54 2003
//  Update count     : 0
//
///////////////////////////////////////////////////////////////////////////////
//
//  Date        Name            Changes/Extensions
//  ----        ----            ------------------
//
///////////////////////////////////////////////////////////////////////////////


#ifndef HeapMin_HH
#define HeapMin_HH



/**
 * This template realises a heap which has the heap condition:
 * The value of fathernode is smaller than each value of son's nodes.
 * To realize it, I chose a field of the datatype T with the size m_numberOfArrays.
 * This size you can influence by setting the parameter m_value.
 *
 * The interface of this template:
 * There are two kinds of Constructors.
 * Destructor.
 * The methods sortIn(int) and heapify(int) to achive the heap condition by 
 * using the method exchange(T*,T*) which exchanges the values of two keys.
 * The method insertKey(T) to enter a new member T to the heap.
 * PrintHeap() is dedicated to get a plot of the hole heap. 
 * To find out whether the heap is empty you can use emptyHeap().
 * By setValue(int) you change the parameter m_value.
 * The method increasNumberOfNodes() increases m_numberOfNodes about one step;
 * equivalent decreasNumberOfNodes() decreases m_numberOfNodes about one step.
 * Last getNumberOfNodes returns m_numberOfNodes.
 *
 * The private part of this template:
 * \param m_numberOfNodes Number of the keys still being involved into the heap.
 * \param m_numberOfArrays Size of the current usable field m_heapAsArray. 
 * \param m_start Saves the number of the first node (always = 1).
 * \param m_value Saves the number of the amount of resizing m_heapAsArray.
 *
 * The protected part of this template:
 * \param T *m_heapAsArray Is a pointer to allocated field m_heapAsArray 
 * from datatype T. 
 */
template <class T>
class HeapMin {

private:
    // Variables:
    int m_numberOfArrays;
    int m_start;
    int m_numberOfNodes;
    int m_value;	

protected:    
    // Pointer to m_heapAsArray 
    T* m_heapAsArray;    	

public:
   /**
    * First there are no keys in the heap, therefore:
    * Set m_numberOfNodes = 0
    * The number of the first node is always 1: set m_start = 1.
    * The default value of field size is 10 : set m_value = 10.
    * The size of the field is given by m_numberOfArrays. 
    * Try {...} tries whether there is enough memory to allocate a field of
    * datatype T with the size m_numberOfArrays. If it's okay, the field will 
    * be allocated and values will be initialized by init. 
    * Otherwise there will be an exception from type bad_alloc and
    * there is no construction.
    * \param init It's the input value to initialize the array. 
    */
    // Constructor  
    HeapMin(T init);

   /**
    * The same features like HeapMin(T init) but further you can set the
    * parameter m_value with your input  _value.If  _value is chosen smaller than 1,
    * there would be no construction with the comment "Invalid input for m_value!" 
    * \param _value It's the input to set the variable m_value. 
    * \param init It's the input value to initialize the array . 
    */
    // Constructor 
    HeapMin(T init,int);

   /**
    * The destructor destroies just the allocated field so that the reserved
    * memory is given free.
    */
    // Destructor
    ~HeapMin();

   /**
    * This method sorts the given node to the right position of the heap.
    * \param node It's the number of the node that should be insert.
    */    
    void sortIn(int);

   /**
    * This method you can call the heart of the whole priority queue: 
    * It keeps or achives the heap condition again (e.g. after extracting the root).
    * This works by comparing the values of father and his sons. If the heap
    * condition is violated, it will be repaired by exchanging the values.
    * After that there is a recurrent call of heapify(selectedSon) until the
    * heap condition is reestablished down to leaves of the heap.
    * \param node It's the number of the node at which may the heap condition
    *             is violated.
    */
    void heapify(int);

   /**
    * This method exchanges the value of two given nodes (call-by-reference).
    * \param T *node1 It's the first given node.
    * \param T *node2 It's the second given node.
    */
    void exchange(T *,T *);



   /**
    * This method adds a new node together with the value  _value to the current
    * heap. First it increases m_numberOfNodes with the amount of 1. After that there
    * runs a check whether the field runs to small through this insertion. If this
    * is the case there will be build up a new field with the new size:
    * old size + m_value. If this allocation is not possible, the variable m_value will
    * be halved and the insertion will be tried again and again until it will work
    * or in bad case (if m_value<50) the program will aborted by exit with the
    * comment "Not enough memory!".
    * If the new allocation was succesfull the new key will be entered at 
    * m_heapAsArray[m_numberOfNodes] (the last node in the heap). Last, the call of 
    * sortIn(m_numberOfNodes) will place it a the right position in the tree.
    * \param _value It's the new value that should be insert.
    */
    void insertKey(T);

   /**
    * This method prints the values of the heap to the screen.
    */
    void printHeap();

   /**
    * Here you are able to check whether the heap is empty or not.
    * \return bool You will get true if the heap is empty and false 
    * if the heap isn't empty. 
    */
    bool emptyHeap();
    
   /**
    * It's an additional function to reset the parameter m_value after 
    * construction.If  _value is chosen smaller than 1, it wouldn't
    * work with the comment "Invalid input for m_value!"
    * \param _value It's the input to re-set the variable m_value.
    */    
    void setValue(int);

   /**
    * This method increases m_numberOfNodes about one step. 
    */
    void increaseNumberOfNodes();

   /**
    * This method decreases m_numberOfNodes about one step. 
    */
    void decreaseNumberOfNodes();

   /**
    * This method returns the current m_numberOfNodes.
    * \return int It's the current m_numberOfNodes.  
    */
    int getNumberOfNodes();
};


// Constructor
template <class T>
inline HeapMin<T>::HeapMin(T init)  {
    // Initializing of the variables
    m_numberOfNodes = 0;
    m_start = 1;
    m_value = 10;
    m_numberOfArrays = m_value;
    // Try to allocate
    try {
    	m_heapAsArray = new T[m_numberOfArrays]; 
    }
    // Not enough memory -> abortion
    catch (std::bad_alloc) {
    	cerr << "Not enough memory!" << endl;
	return;
    }
    // Initializing the allocated field
    for(int i=0;i<=m_numberOfArrays;i=i+1) { 
	m_heapAsArray[i] = init;
    }
}


// Constructor with the possibility to set m_value by _value
template <class T>
inline HeapMin<T>::HeapMin(T init,int _value)  {
    // Check for invalid input
    if(_value<1) {
	cout << "\nInvalid input for m_value\n" ;
	return;
    }
    // After a positive check: Construction 
    else {
	// Initializing of the variables
	m_numberOfNodes = 0;
	m_start = 1;
	m_value = _value;
	m_numberOfArrays = m_value;
        // Try to allocate
	try {
	    m_heapAsArray = new T[m_numberOfArrays]; 
	}
	// Not enough memory -> abortion
	catch (std::bad_alloc) {
	    cerr << "Not enough memory!" << endl;
	    return;
	}
	// Initializing the allocated field
	for(int i=0;i<=m_numberOfArrays;i=i+1) { 
	    m_heapAsArray[i] = init;
	}
    }
}


// Destructor
template <class T>
inline HeapMin<T>::~HeapMin() {
    delete [] m_heapAsArray;
}


// Method to change m_value again after construction
template <class T>
inline void HeapMin<T>::setValue(int _value) {
    // Check for invalid input
    if(_value<1) {
	cout << "\nInvalid input for m_value!\n" ;
	return;
    }
    // After positiv check: 
    else {
	m_value = _value;
    }
}


// Method to increase m_numberOfNodes 
template <class T>
inline void HeapMin<T>::increaseNumberOfNodes() {
    m_numberOfNodes = m_numberOfNodes+1;
}


// Method to decrease m_numberOfNodes
template <class T>
inline void HeapMin<T>::decreaseNumberOfNodes() {
    m_numberOfNodes = m_numberOfNodes-1;
}


// Method to get m_numberOfNodes
template <class T>
inline int HeapMin<T>:: getNumberOfNodes() {
    return m_numberOfNodes;
}


// Method to find out whether the heap is empty or not
template <class T>
inline bool HeapMin<T>::emptyHeap() {
    if (m_numberOfNodes == 0) {
    	return true;
    }
    else {
    	return false;
    }
}


// Prints the complete heap
template <class T>
inline void HeapMin<T>::printHeap() {
    for(int i=m_start; i <= m_numberOfNodes; i = i+1) {
    	cout << m_heapAsArray[i] << " "  ;
    }
    cout << endl ;
}


// Method to add the key _value to the heap
template <class T>
inline void HeapMin<T>::insertKey(T _value) {
        
    // Increase number of nodes
    m_numberOfNodes =m_numberOfNodes+1;
    
    // If the current field m_heapAsArray is to small:
    // Create an new field which is bigger about the size of m_value
    if (m_numberOfNodes >= m_numberOfArrays) {
	m_numberOfArrays = m_numberOfArrays+m_value;  
        T* heapAsArray2;
    	
        // Try to allocate a new field
    	try {
    		 heapAsArray2 = new T[m_numberOfArrays];
   	}
	// Exception handling:
    	catch (std::bad_alloc) {
    	    // If m_value < 50 exit 
	    if(m_value<50) {
		cerr << "Not enough memory!" << endl;
		exit;
	    }
	    // Else halve m_value and try again
	    else {
		m_value = m_value/2;
		insertKey(_value);
	    }
     	}
	
        // Copying from m_heapAsArray to the temporary heapAsArray2
        for (int i=m_start; i<m_numberOfNodes; i=i+1) {  
    	    heapAsArray2[i] = m_heapAsArray[i];
    	}
    	heapAsArray2[m_numberOfNodes] =_value;
    	delete [] m_heapAsArray;
    	m_heapAsArray = heapAsArray2;
    }
    // Write the new key 
    m_heapAsArray[m_numberOfNodes] = _value;
	
    // Sort in the new insert key by calling sortIn
    sortIn(m_numberOfNodes);
}


// Sort in the given node to the right position of the heap
template <class T>
inline void HeapMin<T>::sortIn(int node) {
    if (node>1) {
	// Determine the father node 
    	int parentNumber = (node)/2;
	
	// If the heap condition is violated exchange 
        // the values of father and son by calling the method exchange
        if (m_heapAsArray[node] < m_heapAsArray[parentNumber]) {
	    exchange(&m_heapAsArray[node], &m_heapAsArray[parentNumber]);
	    
	    // Recurent call of sortIn until node>1 
	    sortIn(parentNumber);
        }  
    }
}


// Method to exchange the values of the keys node1 and node2 
template <class T>
inline void HeapMin<T>::exchange(T *node1, T *node2) {
    T temp = *node1;
    *node1 = *node2;
    *node2 = temp;
}


// Method to keep or rather to achive the heap condition
template <class T>
inline void HeapMin<T>::heapify(int node) {
    int leftSon = 2*node;
    int rightSon = 2*node+1 ;
    int selectedSon;
    
    // No existing right son
    if ((leftSon <= m_numberOfNodes) && (rightSon > m_numberOfNodes)) {
    	// If heap condition is violated exchange the value of 
	// father and left son
	if (m_heapAsArray[leftSon] < m_heapAsArray[node]) {
	    exchange(&m_heapAsArray[node],&m_heapAsArray[leftSon]);
        }
    }
    else {
	// Existing right son
	if (rightSon <= m_numberOfNodes) {
	    // Compare of left and right son and choose the smaller once 
	    if (m_heapAsArray[leftSon] < m_heapAsArray[rightSon]) {
		selectedSon = leftSon;
            }
            else {
            	selectedSon = rightSon;
            }
	    // If heap condition is violated exchange the value of 
	    // father and selected son
            if (m_heapAsArray[selectedSon] < m_heapAsArray[node]) {
            	exchange(&m_heapAsArray[node],&m_heapAsArray[selectedSon]);
                heapify(selectedSon);
            }
        }
    }
}
       
#endif



