////////////////////////////////////// c++ ////////////////////////////////////
//
//  Module           : MaxPriorityQueue.hh
//  Description      :
//  Author           : Christoph Laux
//  Email            : claux@ix.urz.uni-heidelberg.de
//  Copyright        : University of Heidelberg
//  Created on       : Thu Febr 27 09:21:54 2003
//  Last modified by : Laux
//  Last modified on : Thu Febr 27 09:21:54 2003
//  Update count     : 0
//
///////////////////////////////////////////////////////////////////////////////
//
//  Date        Name            Changes/Extensions
//  ----        ----            ------------------
//
///////////////////////////////////////////////////////////////////////////////


#ifndef MaxPriorityQueue_HH
#define MaxPriorityQueue_HH




/**
 * This template class realizes a max-priority queue that's based on a heap.
 * Therefore this template class is derived privately from template <class T>
 * class HeapMax, i.e. the biggest value is placed on the root (and the heap
 * condition of HeapMax is obtained). 
 * So the construction is based on a HeapMax<T> and the main methods do already
 * exists in HeapMax.
 * The first two methods are constructors just calling the constructors of
 * HeapMax. There's no further destructor needed - default destructor is okay.
 * The additional methods to realize a priority queue are:
 * extractRoot() To return and cut the root of the heap;
 * deleteKey(int key) To extract the node with number key;
 * changeKey(int key ,T newValue) Here you can change the value of the node key
 * to newValue.
 * The other methods are just calling the corresponding method of HeapMax<T>.
 * You can't see any variables in this template, but you can reach the private
 * variables of template class HeapMax by using the methods:
 *      void setValue(int);
 *	void increaseNumberOfNodes();
 *	void decreaseNumberOfNodes();
 *	int getNumberOfNodes();
 * How they work please read in the descprition of 
 * template <class T>  class HeapMax.
 */
template <class T>
class MaxPriorityQueue : private HeapMax<T> {

public:
    /**
     * Constructor which is calling constructor HeapMax<T>::HeapMax(T init)
     * \param init It's the input value to initialize the array. 
     */
    // Constructor
    MaxPriorityQueue(T init);

    /**
     * Constructor which calls  constructor 
     * HeapMax<T>::HeapMax(T init,int _value)
     * \param init It's the input value to initialize the array. 
     * \param _value It's the input to set the variable m_value. 
     */
    // Constructor
    MaxPriorityQueue(T init,int);
    
    // Default destructor okay

 
    /**
     * Method to insert a new key value by calling 
     * HeapMax<T>::insertKey(T _value).
     * \param _value It's the new value from datatype T that should be insert
     * to the heap.
     */
    void insertKeyInQueue(T);

    /**
     * Method to print the queue by calling void HeapMax<T>::printHeap().
     */
    void printQueue();

    /**
     * Method to check whether the queue is empty or not by calling 
     * HeapMax<T>::emptyHeap().
     * \return bool You get true if the heap is empty and false 
     * if the heap isn't empty. 
     * (ref. HeapMax<T>::emptyHeap())
     */
    bool emptyQueue();

    /**
     * With this method you are able to re-set the variable m_value of HeapMax<T>
     * by calling void HeapMax<T>::setValue(int _value) which is responsible for 
     * the size of the field allocation 
     * (ref. void HeapMax<T>::setValue(int _value)).
     * \param reset It's the future set of the variable m_value.
     */
    void setQueueValue(int);

    /**
     * This method extracts the root of the heap.
     * \return T It's the extracted (and by it deleted) top of the heap. 
     */
    T extractRoot();

    /**
     * Method to delete a member key of the heap.
     * First there runs an area check whether the chosen key is in the admissable
     * area or not. If the check is negative there will be the answer:
     * "The chosen key is out of the admissable area" and there will be nothing
     * more done. If the check is positive the key will be deleted and the heap
     * condition will be reestablished.
     * \param key It's input key that should be deleted.
     */
    void deleteKey(int key);

    /**
     * Method to change the value of a member key to newValue.
     * First there runs an area check whether the chosen key is in the admissable
     * area or not. If the check is negative there will be the answer:
     * "The chosen key is out of the admissable area" and there will be nothing
     * more done.If the check is positive the value of the declared key will be
     * changed to newValue and the heap condition will be reestablished.
     * \param key It's the key of which the value should be changed.
     * \param newValue It's the new to be given value of the declared key.
     */
    void changeKey(int key ,T newValue);
        
};



// Constructor which is calling constructor HeapMax<T>::HeapMax(T init)
template <class T>
inline MaxPriorityQueue<T>::MaxPriorityQueue(T init)
    :HeapMax<T>::HeapMax(init)  {}


// Constructor which is calls constructor HeapMax<T>::HeapMax(T init,int _value)
template <class T>
inline MaxPriorityQueue<T>::MaxPriorityQueue(T init,int _value)
    :HeapMax<T>::HeapMax(init,_value) {}


// This method extracts the root of the heap
template <class T>
inline T MaxPriorityQueue<T>::extractRoot() {
    T root;
    // Reach the variable m_numberOfNodes of HeapMax<T>
    int num = getNumberOfNodes();
    // Assignment of the value of the root to root
    root = m_heapAsArray[1];
    // Assignment of the value of the last key to the root
    m_heapAsArray[1] = m_heapAsArray[num];
    // Cut the last node 
    decreaseNumberOfNodes();
    // Reestablish the heap condition by heapify
    heapify(1);
    return root;
}


// Method to insert a new key by calling void HeapMax<T>::insertKey(T _value)
template <class T>
inline void MaxPriorityQueue<T>::insertKeyInQueue(T value) {
    insertKey(value);
}


// Method to print the queue by calling void HeapMax<T>::printHeap()
template <class T>
inline void MaxPriorityQueue<T>::printQueue() {
    printHeap();
}



// Method to check whether the queue is empty or not by calling
// bool HeapMax<T>::emptyHeap()
template <class T>
inline bool MaxPriorityQueue<T>::emptyQueue() {
    emptyHeap();
}


// With this method you are able to re-set the variable m_value of HeapMax<T>
// by calling void HeapMax<T>::setValue(int _value)
template <class T>
inline void MaxPriorityQueue<T>::setQueueValue(int reset) {
    setValue(reset);
}


// Method to delete a member key of the heap
template <class T>
inline void MaxPriorityQueue<T>::deleteKey(int key) {
    cout << "\nDelete Key Number(" << key << "):" << endl ;
    // Area check 
    if(( key < 1) || (key > getNumberOfNodes())) {
        cerr << "\nThe chosen key "<< key <<" is out of the admissable area\n";
	return;
    }
    // If check is okay:
    else {
	int parent = key/2;
        // Check whether newValue is bigger or smaller than the old value
	// If it's smaller
	if(m_heapAsArray[getNumberOfNodes()] < m_heapAsArray[key]) {
	    exchange(&(m_heapAsArray[key]),&(m_heapAsArray[getNumberOfNodes()]));
	    // Cut the last key(->delete)
	    decreaseNumberOfNodes();
	    // Achive heap condition again
	    heapify(key);
	}
	else {
	    // Otherwise do, while parent!=0, 
	    // (heapify (parent)&& parent=parent/2)
	    m_heapAsArray[key]=m_heapAsArray[getNumberOfNodes()];
	    while(parent != 0) {
		heapify(parent);
		parent = parent/2;
            }
	}
    }
}


// Method to change the value of a member key to newValue 
template <class T>
inline void MaxPriorityQueue<T>::changeKey(int key,T newValue) {
    cout <<"\nChange Key Number("<< key << ") to Value "<< newValue <<":"<< endl;
    // If check is okay:
    if(( key < 1) || (key > getNumberOfNodes())) {
        cerr << "\nThe chosen key "<< key <<" is out of the admissable area\n";
	return;
    }
    // If check is okay:
    else {
	int parent = key/2;
	// Check whether newValue is bigger or smaller than the old value
	// If it's smaller assign and call heapify(key)
	if(newValue < (m_heapAsArray[key])) {
	    m_heapAsArray[key] = newValue;
	    heapify(key);
	}
	// Otherwise assign and do, while parent!=0, 
	// (heapify (parent)&& parent=parent/2)
	else {
	    m_heapAsArray[key] = newValue;
	    while(parent != 0) {
		heapify(parent);
		parent = parent/2;
	    }
	}
    }
}

#endif
